/* Kurtosis1.cpp

Kurtosis RADAR Texture Algorithm v. 1.1
Created by:  Nathan Jennings
             nate@jenningsplanet.com
Created on:  04.15.2010
Updated on:  05.30.2010

Computes Kurtosis Statistical Measure how "peaked" or "flat" a distribution curve is.
A high peak about the mean that falls off rapidly and has "heavy" tails indicate high kurtosis.

From an image processing "texture" point of view large kurtosis values indicates that many pixel values are near the mean
within a local neighborhood (e.g. 3x3, 5x5, 7x7 area) and a smaller number of pixels are further away from the mean
(i.e. are towards the tails of the distribution curve).  One can say that most of the pixels in a local area
are "similar" to one another, but do have some pixels that are very different.  The resulting kurtosis image
can help in determining some land cover types or assit to "segment" the image into areas with distinct texture
patterns.

The majority of this code follows the general logic produced by Ball Aerospace.  The significant contributions
to this code are the statistical measures found below.

 * The information in this file is
 * Copyright(c) 2007 Ball Aerospace & Technologies Corporation
 * and is subject to the terms and conditions of the
 * GNU Lesser General Public License Version 2.1
 * The license text is available from   
 * http://www.gnu.org/licenses/lgpl.html
 */

#include "DataAccessor.h"
#include "DataAccessorImpl.h"
#include "DataRequest.h"
#include "DesktopServices.h"
#include "MessageLogResource.h"
#include "ObjectResource.h"
#include "PlugInArgList.h"
#include "PlugInManagerServices.h"
#include "PlugInRegistration.h"
#include "Progress.h"
#include "RasterDataDescriptor.h"
#include "RasterElement.h"
#include "RasterUtilities.h"
#include "SpatialDataView.h"
#include "SpatialDataWindow.h"
#include "switchOnEncoding.h"
#include "kurtosis1.h"
#include <math.h>   //required for power and square root computations
#include <limits>

REGISTER_PLUGIN_BASIC(RADARTutorial, Kurtosis1);

namespace
{
   template<typename T>
   void Kurtosis(T* pData, DataAccessor pSrcAcc, int row, int col, int rowSize, int colSize)
   {
      
	   double mean = 0;
	   double variance = 0;
	   
	   double kurtosis = 0;
	   double median = 0;
	   double sum = 0;
	   int val = 0;
	   int i=0; //counter for array
	   int j=0;
	   int count = 0;
	   
	   
	   int kernel = 3;             //future enhancements will provide the ability for the use to choose
	                            //window size via a GUI
       short offset = (kernel-1) / 2;
	   int * a = NULL; 
	   a = new int [kernel*2];  //not used yet

	  
	      mean = 0;
		  sum = 0;
		  
		  count = 0;
		  variance = 0;
		  kurtosis = 0;
		     
		     
		     for(i=row-offset ; i <= (row +offset); i++)  //loop over kernel rows
			 {for(j=col-offset; j <= (col +offset); j++)  //loop over kernel cols
		  
		        {  
					pSrcAcc->toPixel(i,j);  //access the specific row,col
					VERIFYNRV(pSrcAcc.isValid());
					count += 1;
                  
           
		  
		   
           T val = *reinterpret_cast<T*>(pSrcAcc->getColumn());
		   

		   sum += val;   //add up the pixels

        

		  } //end of small kernel window  col       
			 } // end of small kernel window row
        
		  
      mean = sum / count; //compute the mean

	     //re-run the loops to compute the variance and kurtosis
          for(i=row-offset ; i <= (row +offset); i++)
			 {for(j=col-offset; j <= (col +offset); j++)
			   {    pSrcAcc->toPixel(i,j);
					VERIFYNRV(pSrcAcc.isValid());
		            T val = *reinterpret_cast<T*>(pSrcAcc->getColumn());
                    variance += pow((val - mean), 2);
					kurtosis += pow((val - mean), 4);
		       }
		     }
  
		     kurtosis = kurtosis / ((kernel*kernel-1) *(variance*variance));	//compute the kurtosis


      *pData = static_cast<T>(kurtosis);
  
	  //May implement some additional scaling mechanisms here. (e.g. log scale with min-max)
	  double* pDoubleData = reinterpret_cast<double*>(pData);
      *pDoubleData = kurtosis;
   }
};

Kurtosis1::Kurtosis1()
{
   setDescriptorId("{90bd7e4d-ca26-4e03-869a-b45dc3856f15}"); //this is unique for Kurtosis
   setName("RADAR Texture Analysis Kurtosis");
   setVersion("Kurtosis v1.1");
   setDescription("Calculate and return a kurtosis texture raster element for a one band "
      "image (raster element).");
   setCreator("Nathan Jennings");
   setCopyright("Copyright (C) 2008, Ball Aerospace & Technologies Corp.; (C) 2010 JenningsPlanet.com");
   setProductionStatus(false);
   setType("Sample");
   setSubtype("Kurtosis");
   setMenuLocation("[RADAR Processing]/Texture Analysis/Kurtosis");
   setAbortSupported(true);
}

Kurtosis1::~Kurtosis1()
{
}

bool Kurtosis1::getInputSpecification(PlugInArgList*& pInArgList)
{
   VERIFY(pInArgList = Service<PlugInManagerServices>()->getPlugInArgList());
   pInArgList->addArg<Progress>(Executable::ProgressArg(), NULL, "Progress reporter");
   pInArgList->addArg<RasterElement>(Executable::DataElementArg(), "Perform kurtosis texture on this data element");
   return true;
}

bool Kurtosis1::getOutputSpecification(PlugInArgList*& pOutArgList)
{
   VERIFY(pOutArgList = Service<PlugInManagerServices>()->getPlugInArgList());
   pOutArgList->addArg<RasterElement>("Result", NULL);
   return true;
}

bool Kurtosis1::execute(PlugInArgList* pInArgList, PlugInArgList* pOutArgList)
{   int kernel = 3;             //future enhancements will provide the ability for the use to choose
	                            //window size via a GUI
       short offset = (kernel-1) / 2;
   //StepResource pStep("Skewness 1", "app", "5EA0CC75-9E0B-4c3d-BA23-6DB7157BBD54");
   StepResource pStep("Kurtosis", "app", "90bd7e4d-ca26-4e03-869a-b45dc3856f15");   //this is unique for kurtosis
   if (pInArgList == NULL || pOutArgList == NULL)
   {
      return false;
   }
   Progress* pProgress = pInArgList->getPlugInArgValue<Progress>(Executable::ProgressArg());
   RasterElement* pCube = pInArgList->getPlugInArgValue<RasterElement>(Executable::DataElementArg());
   if (pCube == NULL)
   {
      std::string msg = "A raster cube must be specified.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }
   RasterDataDescriptor* pDesc = static_cast<RasterDataDescriptor*>(pCube->getDataDescriptor());
   VERIFY(pDesc != NULL);
   if (pDesc->getDataType() == INT4SCOMPLEX || pDesc->getDataType() == FLT8COMPLEX)
   {
      std::string msg = "Kurtosis cannot be performed on complex types.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }

   FactoryResource<DataRequest> pRequest;
   pRequest->setInterleaveFormat(BSQ);
   DataAccessor pSrcAcc = pCube->getDataAccessor(pRequest.release());

   ModelResource<RasterElement> pResultCube(RasterUtilities::createRasterElement(pCube->getName() +
      "_Kurtosis_Result", pDesc->getRowCount(), pDesc->getColumnCount(), FLT8BYTES));  //pDesc->getDataType()
                                                                         //FLT8BYTES required for floating point value
                                                                         //instead of integer
   if (pResultCube.get() == NULL)
   {
      std::string msg = "A raster cube could not be created.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }
   FactoryResource<DataRequest> pResultRequest;
   pResultRequest->setWritable(true);
   DataAccessor pDestAcc = pResultCube->getDataAccessor(pResultRequest.release());

   pDestAcc->getColumn();
   //pDestAcc->nextColumn(1);
   //pDestAcc->nextColumn();

   pDestAcc->getRow();
   pDestAcc->nextRow(offset);  //offsets a row around the output image (i.e. 0s)
   pDestAcc->nextColumn(offset); //offsets a col around the image (i.e. 0s) 


      //processes image data using offset
   for (unsigned int row = offset; row < pDesc->getRowCount()-offset; ++row)
   {
      if (pProgress != NULL)
      {
         pProgress->updateProgress("Calculating result", row * 100 / pDesc->getRowCount(), NORMAL);
      }
      if (isAborted())
      {
         std::string msg = getName() + " has been aborted.";
         pStep->finalize(Message::Abort, msg);
         if (pProgress != NULL)
         {
            pProgress->updateProgress(msg, 0, ABORT);
         }
         return false;
      }
      if (!pDestAcc.isValid())
      {
         std::string msg = "Unable to access the cube data.";
         pStep->finalize(Message::Failure, msg);
         if (pProgress != NULL) 
         {
            pProgress->updateProgress(msg, 0, ERRORS);
         }
         return false;
      }
      for (unsigned int col = offset; col < pDesc->getColumnCount()-offset; ++col)
      {
         switchOnEncoding(pDesc->getDataType(), Kurtosis, pDestAcc->getColumn(), pSrcAcc, row, col, pDesc->getRowCount(), pDesc->getColumnCount());
         pDestAcc->nextColumn();
      }

      pDestAcc->nextRow();
	  pDestAcc->nextColumn();
   }

   if (!isBatch())
   {
      Service<DesktopServices> pDesktop;

      SpatialDataWindow* pWindow = static_cast<SpatialDataWindow*>(pDesktop->createWindow(pResultCube->getName(),
         SPATIAL_DATA_WINDOW));

      SpatialDataView* pView = (pWindow == NULL) ? NULL : pWindow->getSpatialDataView();
      if (pView == NULL)
      {
         std::string msg = "Unable to create view.";
         pStep->finalize(Message::Failure, msg);
         if (pProgress != NULL) 
         {
            pProgress->updateProgress(msg, 0, ERRORS);
         }
         return false;
      }

      pView->setPrimaryRasterElement(pResultCube.get());
      pView->createLayer(RASTER, pResultCube.get());
   }

   if (pProgress != NULL)
   {
      pProgress->updateProgress("Kurtosis is compete.", 100, NORMAL);
   }

   pOutArgList->setPlugInArgValue("Kurtosis_Result", pResultCube.release());

   pStep->finalize();
   return true;
}
